"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformForOfIterableStatement = exports.isIterableExpression = exports.returnsIterableType = exports.isIterableType = void 0;
const ts = require("typescript");
const lua = require("../../../LuaAST");
const extensions = require("../../utils/language-extensions");
const utils_1 = require("../loops/utils");
const variable_declaration_1 = require("../variable-declaration");
const diagnostics_1 = require("../../utils/diagnostics");
const utils_2 = require("../../../utils");
const multi_1 = require("./multi");
function isIterableType(type) {
    return extensions.isExtensionType(type, extensions.ExtensionKind.IterableType);
}
exports.isIterableType = isIterableType;
function returnsIterableType(context, node) {
    const signature = context.checker.getResolvedSignature(node);
    const type = signature === null || signature === void 0 ? void 0 : signature.getReturnType();
    return type ? isIterableType(type) : false;
}
exports.returnsIterableType = returnsIterableType;
function isIterableExpression(context, expression) {
    const type = context.checker.getTypeAtLocation(expression);
    return isIterableType(type);
}
exports.isIterableExpression = isIterableExpression;
function transformForOfMultiIterableStatement(context, statement, block) {
    const luaIterator = context.transformExpression(statement.expression);
    let identifiers = [];
    if (ts.isVariableDeclarationList(statement.initializer)) {
        // Variables declared in for loop
        // for ${initializer} in ${iterable} do
        const binding = utils_1.getVariableDeclarationBinding(context, statement.initializer);
        if (ts.isArrayBindingPattern(binding)) {
            identifiers = binding.elements.map(e => variable_declaration_1.transformArrayBindingElement(context, e));
        }
        else {
            context.diagnostics.push(diagnostics_1.invalidMultiIterableWithoutDestructuring(binding));
        }
    }
    else if (ts.isArrayLiteralExpression(statement.initializer)) {
        // Variables NOT declared in for loop - catch iterator values in temps and assign
        // for ____value0 in ${iterable} do
        //     ${initializer} = ____value0
        identifiers = statement.initializer.elements.map((_, i) => lua.createIdentifier(`____value${i}`));
        if (identifiers.length > 0) {
            block.statements.unshift(lua.createAssignmentStatement(statement.initializer.elements.map(e => utils_2.cast(context.transformExpression(e), lua.isAssignmentLeftHandSideExpression)), identifiers));
        }
    }
    else {
        context.diagnostics.push(diagnostics_1.invalidMultiIterableWithoutDestructuring(statement.initializer));
    }
    if (identifiers.length === 0) {
        identifiers.push(lua.createAnonymousIdentifier());
    }
    return lua.createForInStatement(block, identifiers, [luaIterator], statement);
}
function transformForOfIterableStatement(context, statement, block) {
    var _a;
    const type = context.checker.getTypeAtLocation(statement.expression);
    if (((_a = type.aliasTypeArguments) === null || _a === void 0 ? void 0 : _a.length) === 2 && multi_1.isMultiReturnType(type.aliasTypeArguments[0])) {
        return transformForOfMultiIterableStatement(context, statement, block);
    }
    const luaIterator = context.transformExpression(statement.expression);
    const identifier = utils_1.transformForInitializer(context, statement.initializer, block);
    return lua.createForInStatement(block, [identifier], [luaIterator], statement);
}
exports.transformForOfIterableStatement = transformForOfIterableStatement;
//# sourceMappingURL=iterable.js.map