"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformReturnStatement = exports.transformExpressionBodyToReturnStatement = void 0;
const ts = require("typescript");
const lua = require("../../LuaAST");
const annotations_1 = require("../utils/annotations");
const assignment_validation_1 = require("../utils/assignment-validation");
const lua_ast_1 = require("../utils/lua-ast");
const scope_1 = require("../utils/scope");
const typescript_1 = require("../utils/typescript");
const call_1 = require("./call");
const multi_1 = require("./language-extensions/multi");
const diagnostics_1 = require("../utils/diagnostics");
function transformExpressionsInReturn(context, node, insideTryCatch) {
    const expressionType = context.checker.getTypeAtLocation(node);
    if (ts.isCallExpression(node)) {
        // $multi(...)
        if (multi_1.isMultiFunctionCall(context, node)) {
            // Don't allow $multi to be implicitly cast to something other than LuaMultiReturn
            const type = context.checker.getContextualType(node);
            if (type && !multi_1.isMultiReturnType(type)) {
                context.diagnostics.push(diagnostics_1.invalidMultiFunctionReturnType(node));
            }
            let returnValues = call_1.transformArguments(context, node.arguments);
            if (insideTryCatch) {
                returnValues = [lua_ast_1.wrapInTable(...returnValues)]; // Wrap results when returning inside try/catch
            }
            return returnValues;
        }
        // Force-wrap LuaMultiReturn when returning inside try/catch
        if (insideTryCatch && multi_1.returnsMultiType(context, node) && !multi_1.shouldMultiReturnCallBeWrapped(context, node)) {
            return [lua_ast_1.wrapInTable(context.transformExpression(node))];
        }
    }
    else if (multi_1.isInMultiReturnFunction(context, node) && multi_1.isMultiReturnType(expressionType)) {
        // Unpack objects typed as LuaMultiReturn
        return [lua_ast_1.createUnpackCall(context, context.transformExpression(node), node)];
    }
    if (!annotations_1.isInTupleReturnFunction(context, node)) {
        return [context.transformExpression(node)];
    }
    let results;
    // Parent function is a TupleReturn function
    if (ts.isArrayLiteralExpression(node)) {
        // If return expression is an array literal, leave out brackets.
        results = node.elements.map(e => context.transformExpression(e));
    }
    else if (!annotations_1.isTupleReturnCall(context, node) && typescript_1.isArrayType(context, expressionType)) {
        // If return expression is an array-type and not another TupleReturn call, unpack it
        results = [lua_ast_1.createUnpackCall(context, context.transformExpression(node), node)];
    }
    else {
        results = [context.transformExpression(node)];
    }
    // Wrap tupleReturn results when returning inside try/catch
    if (insideTryCatch) {
        results = [lua_ast_1.wrapInTable(...results)];
    }
    return results;
}
function transformExpressionBodyToReturnStatement(context, node) {
    const expressions = transformExpressionsInReturn(context, node, false);
    return lua.createReturnStatement(expressions, node);
}
exports.transformExpressionBodyToReturnStatement = transformExpressionBodyToReturnStatement;
const transformReturnStatement = (statement, context) => {
    // Bubble up explicit return flag and check if we're inside a try/catch block
    let insideTryCatch = false;
    for (const scope of scope_1.walkScopesUp(context)) {
        scope.functionReturned = true;
        if (scope.type === scope_1.ScopeType.Function) {
            break;
        }
        insideTryCatch = insideTryCatch || scope.type === scope_1.ScopeType.Try || scope.type === scope_1.ScopeType.Catch;
    }
    let results;
    if (statement.expression) {
        const expressionType = context.checker.getTypeAtLocation(statement.expression);
        const returnType = context.checker.getContextualType(statement.expression);
        if (returnType) {
            assignment_validation_1.validateAssignment(context, statement, expressionType, returnType);
        }
        results = transformExpressionsInReturn(context, statement.expression, insideTryCatch);
    }
    else {
        // Empty return
        results = [];
    }
    if (insideTryCatch) {
        results.unshift(lua.createBooleanLiteral(true));
    }
    return lua.createReturnStatement(results, statement);
};
exports.transformReturnStatement = transformReturnStatement;
//# sourceMappingURL=return.js.map